package org.artifactory.servicetrust;

import org.apache.commons.lang.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.access.GenericTokenSpec;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.util.CollectionUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.common.JsonUtils;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Stream;

import static org.jfrog.access.common.AccessAuthz.ADMIN;

/**
 * @author volodymyrm
 */
public class ArtifactoryColdTokenSpecService implements TokenSpecService {

    private static final String LIMITED_SCOPE = "scoped";
    public static final ArtifactoryColdTokenSpecService ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE =
            new ArtifactoryColdTokenSpecService();

    @Override
    public TokenSpec<GenericTokenSpec> createPairingTokenSpec(@NotNull ServiceId artifactoryServiceId,
            @NotNull ServiceTrust trust, @NotNull String subject, @NotNull String scope, @NotNull String baseUrl,
            @Nullable String pairingUrl, @NotNull String exchangeUrl) {
        if (trust.equals(ServiceTrust.TWO_WAY)) {
            throw new IllegalArgumentException("artifactory-cold does not support two-way pairing");
        }
        verifyIsColdStorageEnabled();
        String useCasePath = getUseCaseFromPairScope(Collections.singletonList(scope));
        if (ServiceTrustUseCase.ARTIFACTORY_COLD.matches(useCasePath)) {
            useCasePath = generateIceUseCaseWithNamespace(useCasePath);
        }
        String updatedScope = String.format(SCOPE_PATTERN, useCasePath);
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.audience(Collections.singletonList(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS);
        spec.scope(Collections.singletonList(updatedScope));
        spec.refreshable(false);
        TokenExtension tokenExtension = new TokenExtension(baseUrl, exchangeUrl, null, true);
        String extension = JsonUtils.getInstance().valueToString(tokenExtension);
        spec.extension(extension);
        return spec;
    }

    @Override
    public TokenSpec<GenericTokenSpec> createMasterTokenSpec(@NotNull ServiceId artifactoryServiceId,
            @NotNull String useCasePath, @NotNull String pairingScope, @NotNull String baseUrl,
            @NotNull String exchangeUrl) {
        verifyIsColdStorageEnabled();
        String subject = artifactoryServiceId.getFormattedName();
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.audience(Collections.singletonList(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_MASTER_TOKEN_EXPIRATION_TIME_IN_SECS);
        String adminScope = artifactoryServiceId.getFormattedName() + ":" + ADMIN;
        String requiredScope = INTERNAL_SCOPE_PREFIX + useCasePath;
        spec.scope(Arrays.asList(adminScope, requiredScope, pairingScope));
        spec.refreshable(true);
        TokenExtension newExtension = new TokenExtension(baseUrl, exchangeUrl, null, true);
        String extension = JsonUtils.getInstance().valueToString(newExtension);
        spec.extension(extension);
        return spec;
    }

    @Override
    public TokenSpec<GenericTokenSpec> createExchangeTokenSpec(ServiceId artifactoryServiceId, String useCasePath) {
        verifyIsColdStorageEnabled();
        String subject = artifactoryServiceId.getFormattedName();
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.audience(Collections.singletonList(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_EXCHANGE_TOKEN_EXPIRATION_TIME_IN_SECS);
        String adminScope = artifactoryServiceId.getFormattedName() + ":" + ADMIN;
        String requiredScope = INTERNAL_SCOPE_PREFIX + useCasePath;
        spec.scope(Arrays.asList(adminScope, requiredScope));
        spec.refreshable(false);
        String extension = JsonUtils.getInstance().valueToString(TokenExtension.REVOCABLE_TOKEN);
        spec.extension(extension);
        return spec;
    }

    @Override
    public ExchangeTokenRequest createExchangeTokenRequest(String subject, String useCasePath, JwtAccessToken jwt) {
        verifyIsWarmStorageEnabled();
        String updatedUseCasePath = getUseCaseFromPairScope(jwt.getScope());
        return new ExchangeTokenRequest(subject, updatedUseCasePath);
    }

    @Override
    public void storeMasterTokenData(ExchangeTokenResponse response, TokenExtension extension) {
        verifyIsWarmStorageEnabled();
        if (StringUtils.isEmpty(response.getScope())) {
            throw new IllegalArgumentException("Master token does not have scope");
        }
        String[] tokens = response.getScope().split(" ");
        String tokenWithNamespace = findFirstRetentionToken(Arrays.stream(tokens));
        if (tokenWithNamespace == null) {
            throw new IllegalArgumentException("Master token scope with the 'artifactory-cold' scope does not exist");
        }
        String namespace = tokenWithNamespace.replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX, "")
                .replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX, "")
                .replace(ServiceTrustUseCase.ARTIFACTORY_COLD.getName() + "/", "")
                .replace(ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), "")
                .replace(INTERNAL_SCOPE_PREFIX, "");

        if (StringUtils.isEmpty(namespace)) {
            throw new IllegalArgumentException("Can not parse the master token scope to get a cold namespace: " +
                    "namespace is empty");
        }
        ArtifactRetentionAddon artifactRetentionAddon = ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
        artifactRetentionAddon.saveNamespaceAndMasterToken(extension.getBaseUrl(), extension.getExchangeUrl(),
                namespace, response.getTokenId(), response.getAccessToken(), response.getTokenType(),
                response.getExpiresIn(), response.getScope());
    }

    @Override
    public void updateMasterTokenData(ExchangeTokenResponse response) {
        verifyIsWarmStorageEnabled();
        ArtifactRetentionAddon artifactRetentionAddon = ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
        artifactRetentionAddon.updateMasterToken(response.getTokenId(), response.getAccessToken(),
                response.getTokenType(), response.getExpiresIn(), response.getScope());
    }

    @Override
    public ExchangeTokenRequest createMasterTokenExchangeRequest(@NotNull String subject) {
        verifyIsWarmStorageEnabled();
        ArtifactRetentionAddon artifactRetentionAddon = ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
        String namespace = artifactRetentionAddon.getJfrogColdStorageArchivingNamespace();
        if (StringUtils.isEmpty(namespace)) {
            throw new IllegalStateException("Can not create exchange request to refresh master token: " +
                    "namespace does not exist");
        }
        return new ExchangeTokenRequest(subject,
                String.join("/", ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), namespace));
    }

    @Override
    public ExchangeTokenRequest createScopedTokenExchangeRequest(@NotNull String subject) {
        verifyIsWarmStorageEnabled();
        ArtifactRetentionAddon artifactRetentionAddon = ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
        String namespace = artifactRetentionAddon.getJfrogColdStorageArchivingNamespace();
        if (StringUtils.isEmpty(namespace)) {
            throw new ServiceTrustException("Can not create exchange request to refresh master token: " +
                    "namespace does not exist");
        }
        String target =
                String.format("%s/%s:%s", ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), namespace, LIMITED_SCOPE);
        return new ExchangeTokenRequest(subject, target);
    }

    /**
     * Extract a use case with namespace 'artifactory-cold/namespace' from scopes
     */
    @NotNull
    private String getUseCaseFromPairScope(List<String> scope) {
        if (CollectionUtils.isNullOrEmpty(scope)) {
            throw new IllegalArgumentException("Pair's token scope can not be null");
        }
        String retentionScope = findFirstRetentionToken(scope.stream());
        if (retentionScope == null) {
            throw new IllegalArgumentException("Pair's token does not have retention related scope");
        }
        return retentionScope.replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX, "")
                .replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX, "");
    }

    private String findFirstRetentionToken(@NotNull Stream<String> tokens) {
        return tokens
                .filter(t -> t.contains(ServiceTrustUseCase.ARTIFACTORY_COLD.getName()))
                .findFirst()
                .orElse(null);
    }

    private void verifyIsColdStorageEnabled() {
        ArtifactRetentionAddon artifactRetentionAddon = ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
        if (!artifactRetentionAddon.isColdInstance()) {
            throw new IllegalStateException("Unable to execute an action: cold mode is disabled");
        }
    }

    private void verifyIsWarmStorageEnabled() {
        ArtifactRetentionAddon artifactRetentionAddon = ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
        if (artifactRetentionAddon.isColdInstance()) {
            throw new IllegalStateException("Unable to execute an action: instance running in the cold mode");
        }
    }

    /**
     * Generates use case with namespace for the retention feature. It should look like 'artifactory-cold/namespace'
     */
    @NotNull
    private String generateIceUseCaseWithNamespace(@NotNull String useCaseWithoutNamespace) {
        ArtifactRetentionAddon artifactRetentionAddon = ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
        String namespace = artifactRetentionAddon.generateJFrogColdStorageNamespace();
        return String.join("/", useCaseWithoutNamespace, namespace);
    }
}