/*
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2019 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.artifactory.common.home;

import com.jfrog.sysconf.SysConfigConstants;
import com.jfrog.sysconf.SysLayout;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.test.TestUtils;
import org.mockito.Mockito;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.List;
import java.util.stream.Collectors;

import static org.artifactory.common.home.ArtifactoryHome.FORCE_DUPLICATE_REPLICATION_CONVERT_MARKER;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Yossi Shaul
 */
@Test
public class ArtifactoryHomeTest {

    @AfterMethod
    public void cleanSysProps() {
        final List<String> jfProps = System.getProperties().keySet().stream()
                .map(String::valueOf)
                .filter(propKey -> propKey.startsWith("jf."))
                .collect(Collectors.toList());
        jfProps.forEach(System::clearProperty);
    }

    public void testNoHaConfigured() {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        ArtifactoryHome home = new ArtifactoryHome(homeDir.getAbsolutePath());
        assertFalse(home.isHaConfigured());
    }

    public void testHaConfigured() {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        System.setProperty("jf.shared.node.haEnabled", "true");
        ArtifactoryHome home = new ArtifactoryHome(homeDir.getAbsolutePath());
        assertTrue(home.isHaConfigured());
    }

    public void readHomeFolderFromSysProperty() {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        System.setProperty(SysConfigConstants.SYS_JF_PRODUCT_HOME, homeDir.getAbsolutePath());
        ArtifactoryHome home = new ArtifactoryHome(null);
        assertEquals(home.getHomeDir().getAbsolutePath(), homeDir.getAbsolutePath());
    }

    public void testFolderLocations() {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        SysLayout layout = new SysLayout(homeDir, ArtifactoryHome.ARTIFACTORY_SERVICE_NAME);
        System.setProperty(SysConfigConstants.SYS_JF_PRODUCT_HOME, homeDir.getAbsolutePath());

        ArtifactoryHome home = new ArtifactoryHome(null);
        assertEquals(home.getHomeDir().getAbsolutePath(), homeDir.getAbsolutePath());
        assertEquals(home.getEtcDir().getAbsolutePath(), layout.getServiceEtc().getAbsolutePath());
        assertEquals(home.getSecurityDir().getAbsolutePath(),
                new File(layout.getServiceEtc(), SysLayout.SECURITY_DIR).getAbsolutePath());
        assertEquals(home.getDataDir().getAbsolutePath(), layout.getServiceData().getAbsolutePath());
        assertEquals(home.getBackupDir().getAbsolutePath(), layout.getServiceBackup().getAbsolutePath());
        assertEquals(home.getLogDir().getAbsolutePath(), layout.getServiceLog().getAbsolutePath());
        assertEquals(home.getTempWorkDir().getAbsolutePath(),
                Paths.get(layout.getServiceData().getAbsolutePath(), "tmp", "work").toString());
    }

    public void testCustomHaDirs() {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        final String haDataDir = Paths.get(homeDir.getAbsolutePath(), "ha-data").toString();
        final String haBackupDir = Paths.get(homeDir.getAbsolutePath(), "ha-backup").toString();

        System.setProperty(SysConfigConstants.SYS_JF_PRODUCT_HOME, homeDir.getAbsolutePath());
        System.setProperty("jf.shared.node.contextUrl", "http://this.will.enable.ha");
        System.setProperty("jf.artifactory.node.haDataDir", haDataDir);
        System.setProperty("jf.artifactory.node.haDataBackup", haBackupDir);
        ArtifactoryHome home = new ArtifactoryHome(null);

        assertEquals(home.getDataDir().getAbsolutePath(), haDataDir);
        assertEquals(home.getBackupDir().getAbsolutePath(), haBackupDir);
    }

    public void testCustomCacheDir() {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        ArtifactoryHome home = Mockito.spy(new ArtifactoryHome(homeDir.getAbsolutePath()));
        ArtifactorySystemProperties systemProperties = Mockito.mock(ArtifactorySystemProperties.class);
        Mockito.when(home.getArtifactoryProperties()).thenReturn(systemProperties);
        ArtifactoryHome.bind(home);
        try {
            assertEquals(home.getCacheDir().getAbsolutePath(),
                    Paths.get(home.getDataDir().getAbsolutePath(), ".cache").toString());

            String dotCachePath = "/home/dir/dot/cache";
            Mockito.when(systemProperties.getProperty(ConstantValues.dotCacheMetadataFolder)).thenReturn(
                    dotCachePath);
            assertEquals(home.getCacheDir().getAbsolutePath(),
                    dotCachePath);

            dotCachePath = "/home/dir/dot/cache/";
            Mockito.when(systemProperties.getProperty(ConstantValues.dotCacheMetadataFolder)).thenReturn(
                    dotCachePath);
            assertEquals(home.getCacheDir().getAbsolutePath(),
                    dotCachePath.substring(0, dotCachePath.lastIndexOf("/")));
        } finally {
            ArtifactoryHome.unbind();
        }
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = ".*only supported " +
            "for 'data' and 'backup'.*")
    public void testForbiddenCustomHaDir() throws IOException {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        System.setProperty(SysConfigConstants.SYS_JF_PRODUCT_HOME, homeDir.getAbsolutePath());
        ArtifactoryHome home = new ArtifactoryHome(null);
        home.getHAAwareSubDir("nono", new File(homeDir, "ha-nono"));
    }

    @Test
    public void testGetForceDuplicateReplicationConvertFile() {
        File homeDir = TestUtils.createTempDir(ArtifactoryHomeTest.class);
        homeDir.deleteOnExit();
        ArtifactoryHome home = new ArtifactoryHome(homeDir.getAbsolutePath());
        assertEquals(home.getForceDuplicateReplicationConvertFile().getAbsolutePath(),
                Paths.get(home.getEtcDir().getAbsolutePath(), FORCE_DUPLICATE_REPLICATION_CONVERT_MARKER).toFile()
                        .getAbsolutePath());
    }
}