package org.artifactory.rest.common.validator;

import com.google.common.collect.Lists;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.rest.common.exception.RepoConfigException;
import org.artifactory.util.UiRequestUtils;
import org.jdom2.Verifier;

import java.util.List;
import java.util.Optional;

import static org.apache.commons.lang.StringUtils.isBlank;
import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.isSystemRepo;

public class RepositoryNameValidator {

    private static final int REPO_KEY_MAX_LENGTH = 64;
    private static final List<String> forbiddenNames = List.of(".", "..", "&");
    private static final List<Character> forbiddenChars = Lists
            .newArrayList('/', '\\', ':', '|', '?', '*', '"', '<', '>');
    private static final List<String> blacklistPrefixes = List.of("artifactory-", "jfrog-");

    public static void validateRepoName(String repoKey, String repoType, CentralConfigService centralConfig)
            throws RepoConfigException {
        isKeyBlank(repoKey);
        isMaxLengthExceed(repoKey, repoType);
        isReservedName(repoKey);
        isBlacklistedPrefix(repoKey);
        isForbiddenName(repoKey);
        isForbiddenCharacter(repoKey);
        isErrorInXML(repoKey);
        isKeyAvailable(repoKey, centralConfig);
    }

    private static void isKeyBlank(String repoKey) throws RepoConfigException {
        if (isBlank(repoKey)) {
            throw new RepoConfigException("Repository key cannot be empty", SC_BAD_REQUEST);
        }
    }

    private static void isMaxLengthExceed(String repoKey, String repoType) throws RepoConfigException {
        //Local repos limited at 64, remote repos append '-cache' to cache repos which is 6 chars.
        int maxLength = "remote".equals(repoType) ? (REPO_KEY_MAX_LENGTH - 6) : REPO_KEY_MAX_LENGTH;
        if (StringUtils.length(repoKey) > maxLength) {
            throw new RepoConfigException("Repository key exceeds maximum length (" + maxLength + ") chars.",
                    SC_BAD_REQUEST);
        }
    }

    private static void isReservedName(String repoKey) throws RepoConfigException {
        if (UiRequestUtils.isReservedName(repoKey) || isSystemRepo(repoKey)) {
            throw new RepoConfigException("Repository key '" + repoKey + "' is a reserved name", SC_BAD_REQUEST);
        }
    }

    private static void isBlacklistedPrefix(String repoKey) throws RepoConfigException {
        Optional<String> prefix = blacklistPrefixes.stream().filter(repoKey::startsWith).findFirst();
        if (prefix.isPresent()) {
            throw new RepoConfigException(
                    "Repository key '" + repoKey + "' contains blacklisted prefix '" + prefix.get() + "'",
                    SC_BAD_REQUEST);
        }
    }

    private static void isForbiddenName(String repoKey) throws RepoConfigException {
        if (forbiddenNames.stream().anyMatch(repoKey::equals)) {
            throw new RepoConfigException("Invalid Repository key", SC_BAD_REQUEST);
        }
    }

    private static void isForbiddenCharacter(String repoKey) throws RepoConfigException {
        if (repoKey.chars().anyMatch(chr -> forbiddenChars.contains((char) chr))) {
            throw new RepoConfigException("Illegal Repository key : '/,\\,:,|,?,<,>,*,\"' is not allowed",
                    SC_BAD_REQUEST);
        }
    }

    private static void isErrorInXML(String repoKey) throws RepoConfigException {
        String error = Verifier.checkXMLName(repoKey);
        if (StringUtils.isNotBlank(error)) {
            throw new RepoConfigException("Repository key contains illegal character", SC_BAD_REQUEST);
        }
    }

    private static void isKeyAvailable(String repoKey, CentralConfigService centralConfigService)
            throws RepoConfigException {
        if (!centralConfigService.getMutableDescriptor().isKeyAvailable(repoKey)) {
            throw new RepoConfigException("Case insensitive repository key already exists", SC_BAD_REQUEST);
        }
    }
}
