package org.artifactory.ui.rest.service.utils.repositories;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.utils.repositories.RepoKeyType;
import org.artifactory.ui.rest.model.utils.repositories.RepositoriesData;
import org.junit.Before;
import org.junit.Test;
import org.junit.experimental.runners.Enclosed;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.collections.Lists;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.artifactory.repo.system.SystemRepos.LOG_SHIPPING_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.USAGE_REPO_KEY;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertTrue;

@RunWith(Enclosed.class)
public class GetAllRepositoriesServiceTest {

    private static LocalRepoDescriptor createLocalRepoDescriptor(String repoKey) {
        return createLocalRepoDescriptor(repoKey, RepoType.Alpine);
    }

    private static LocalRepoDescriptor createLocalRepoDescriptor(String repoKey, RepoType type) {
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(type);
        localRepoDescriptor.setKey(repoKey);
        return localRepoDescriptor;
    }

    private static FederatedRepoDescriptor createFederatedRepoDescriptor(String repoKey) {
        return createFederatedRepoDescriptor(repoKey, RepoType.Docker);
    }

    private static FederatedRepoDescriptor createFederatedRepoDescriptor(String repoKey, RepoType type) {
        FederatedRepoDescriptor federatedRepoDesc = new FederatedRepoDescriptor();
        federatedRepoDesc.setType(type);
        federatedRepoDesc.setKey(repoKey);
        return federatedRepoDesc;
    }

    @RunWith(Parameterized.class)
    public static class ParameterizedTests {

        private static final String PROJECT_A = "projectA";
        private static final String LOCAL_A = "local-repo";
        private static final String LOCAL_PROJ = "prj-local-repo";
        private final String action;
        private final String listTypeToTest;
        private final int repoCountOnDefault;
        @Mock
        private AuthorizationService authorizationService;
        @Mock
        private RepositoryService repoService;
        @Mock
        private CentralConfigService centralConfigService;
        @Mock
        private ProjectsService projectsService;
        @Mock
        private ArtifactoryRestRequest request;
        private ArtifactoryRestResponse response;
        private GetAllRepositoriesService getAllRepositoriesService;

        public ParameterizedTests(String action, String listTypeToTest, int repoCountOnDefault) {
            this.action = action;
            this.listTypeToTest = listTypeToTest;
            this.repoCountOnDefault = repoCountOnDefault;
        }

        @Parameterized.Parameters(name = "testExecuteWithDifferentActions: {0}")
        public static Collection provideAction() {
            return Arrays.asList(new Object[][] {
                    {"search", "repoTypesList", 2},
                    {"deploy", "repoTypesList", 2},
                    {"packageSearch", "repoTypesList", 2},
                    {"user", "repoTypesList", 3},
                    {"local", "repoList", 2},
                    {"permission", "repoTypesList", 2},
                    {"backup", "repoList", 2},
            });
        }

        @Before
        public void setUp() {
            MockitoAnnotations.openMocks(this);
            response = new ArtifactoryRestResponse();
            getAllRepositoriesService = Mockito.spy(new GetAllRepositoriesService());
            getAllRepositoriesService.setAuthorizationService(authorizationService);
            getAllRepositoriesService.setRepoService(repoService);
            getAllRepositoriesService.setCentralConfigService(centralConfigService);
            getAllRepositoriesService.setProjectsService(projectsService);

            final Set<String> projectRepos = Set.of(LOCAL_PROJ);
            when(authorizationService.canRead(any())).thenReturn(true);
            when(projectsService.getProjectRepositoriesKeys(PROJECT_A, false)).thenReturn(projectRepos);
            when(repoService.getLocalFederatedAndCachedRepoDescriptorsByProject(ProjectsService.DEFAULT_PROJECT))
                    .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_A),
                            createLocalRepoDescriptor(LOCAL_PROJ)));
            when(repoService.getLocalFederatedAndCachedRepoDescriptorsByProject(PROJECT_A))
                    .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_PROJ),
                            createLocalRepoDescriptor(LOG_SHIPPING_REPO_KEY, RepoType.Generic),
                            createLocalRepoDescriptor(USAGE_REPO_KEY, RepoType.Generic)));
            when(repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(PROJECT_A))
                    .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_PROJ),
                            createLocalRepoDescriptor("pipe-info", RepoType.PipeInfo),
                            createLocalRepoDescriptor(LOG_SHIPPING_REPO_KEY, RepoType.Generic),
                            createLocalRepoDescriptor(USAGE_REPO_KEY, RepoType.Generic)));
            when(repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(ProjectsService.DEFAULT_PROJECT))
                    .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_A),
                            createLocalRepoDescriptor(LOCAL_PROJ),
                            createLocalRepoDescriptor(LOG_SHIPPING_REPO_KEY, RepoType.Generic),
                            createLocalRepoDescriptor(USAGE_REPO_KEY, RepoType.Generic)));
            when(repoService.isInfoRepo(RepoType.PipeInfo)).thenReturn(true);
            when(authorizationService.isAdmin()).thenReturn(true);
            MutableCentralConfigDescriptor configDescriptor = new CentralConfigDescriptorImpl();
            configDescriptor.setFileUploadMaxSizeMb(1);
            configDescriptor.setLocalRepositoriesMap(Map.of(LOCAL_A, createLocalRepoDescriptor(LOCAL_A), LOCAL_PROJ,
                    createLocalRepoDescriptor(LOCAL_PROJ)));
            when(centralConfigService.getMutableDescriptor()).thenReturn(configDescriptor);
            when(centralConfigService.getDescriptor()).thenReturn(configDescriptor);
        }

        @Test
        public void testExecuteWithDifferentActions() {
            when(request.getQueryParamByKey(action)).thenReturn("true");

            // Default
            when(request.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn("");
            getAllRepositoriesService.execute(request, response);
            RepositoriesData repoData = (RepositoriesData) response.getIModel();
            if ("repoTypesList".equals(listTypeToTest)) {
                assertEquals(repoCountOnDefault, repoData.getRepoTypesList().size());
                assertNull(repoData.getRepoList());
            } else {
                assertEquals(repoCountOnDefault, repoData.getRepoList().size());
                assertNull(repoData.getRepoTypesList());
            }

            // Project
            when(request.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn(PROJECT_A);
            response = new ArtifactoryRestResponse();
            getAllRepositoriesService.execute(request, response);
            repoData = (RepositoriesData) response.getIModel();
            if ("repoTypesList".equals(listTypeToTest)) {
                if (action.equals("search")) {
                    assertEquals(2, repoData.getRepoTypesList().size());
                } else {
                    assertEquals(1, repoData.getRepoTypesList().size());
                }
                assertNull(repoData.getRepoList());
            } else {
                assertEquals(1, repoData.getRepoList().size());
                assertNull(repoData.getRepoTypesList());
            }
        }
    }

    public static class nonParametrized {

        @Mock
        private CentralConfigService centralConfigService;
        @Mock
        private AuthorizationService authorizationService;

        private GetAllRepositoriesService getAllRepositoriesService;

        @Before
        public void setUp() {
            MockitoAnnotations.openMocks(this);
            getAllRepositoriesService = Mockito.spy(new GetAllRepositoriesService());
            getAllRepositoriesService.setCentralConfigService(centralConfigService);
            getAllRepositoriesService.setAuthorizationService(authorizationService);
        }

        @Test
        public void getRepoDataForUserTest() {
            // setup
            when(authorizationService.isAdmin()).thenReturn(true);

            String alpineLocal = "alpine-local";
            String fedA = "federated-a";
            String fedB = "federated-b";

            RepositoriesData repositoriesData = new RepositoriesData();
            repositoriesData.setRepoList(List.of(alpineLocal, fedA));
            List<String> repoData = List.of(alpineLocal, fedA, fedB);
            doReturn(repoData).when(getAllRepositoriesService).getRepoData(true, true, "");

            CentralConfigDescriptorImpl centralConfigDescriptor = new CentralConfigDescriptorImpl();
            // add locals
            centralConfigDescriptor.getLocalRepositoriesMap().put(alpineLocal, createLocalRepoDescriptor(alpineLocal));
            // add federated
            Map<String, FederatedRepoDescriptor> federatedRepositoriesMap =
                    centralConfigDescriptor.getFederatedRepositoriesMap();
            federatedRepositoriesMap.put(fedA, createFederatedRepoDescriptor(fedA));
            federatedRepositoriesMap.put(fedA, createFederatedRepoDescriptor(fedA));
            federatedRepositoriesMap.put(fedB, createFederatedRepoDescriptor(fedB, RepoType.Npm));
            when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);

            // act
            getAllRepositoriesService.getRepoDataForUser(repositoriesData, "");

            // assert
            List<RepoKeyType> repoTypesList = repositoriesData.getRepoTypesList();
            assertEquals(3, repositoriesData.getRepoTypesList().size());
            assertTrue(repoTypesList.get(0).toString().contains(alpineLocal));
            assertTrue(repoTypesList.get(1).toString().contains(fedA));
            assertTrue(repoTypesList.get(2).toString().contains(fedB));
        }
    }
}